from pathlib import Path
import logging as lg

import click

from prismarine.prisma_common import set_path
from prismarine.prisma_client import generate_client


@click.group()
@click.pass_context
@click.option(
    '--base', required=True, type=click.Path(exists=True),
    help="Primary Python path to use while searching for 'models' package"
)
@click.option(
    '--path', multiple=True, type=click.Path(exists=True),
    help='Additional Python path to use while generating the client'
)
@click.option(
    '--runtime', required=False,
    help="Cluster' package to use in runtime. If not provided, assume top-level"
)
@click.option(
    '--dynamo-access-module', required=False,
    help='''
Dynamo access module to use in runtime. If not provided, DefaultDynamoAccess access class from prismarine.runtime.dynamo_default will be used''',
)
@click.option(
    '--verbose', is_flag=True, help='Enable verbose logging'
)
def prismarine(ctx, path, base, runtime, verbose, dynamo_access_module):
    ctx.ensure_object(dict)
    lg.basicConfig(level=lg.DEBUG if verbose else lg.INFO)
    obj = ctx.obj
    paths = list(Path(p).resolve() for p in path) if path else []
    obj['BaseDir'] = Path(base)
    obj['Runtime'] = runtime
    obj['DynamoAccessModule'] = dynamo_access_module
    set_path(paths)


def main():
    prismarine.add_command(generate_client)
    prismarine()


if __name__ == '__main__':
    main()



from types_boto3_dynamodb import ServiceResource
from types_boto3_dynamodb.service_resource import Table


class DynamoAccess:
    def get_resource(self) -> ServiceResource: ...
    def get_table(self, full_model_name: str) -> Table: ...



import importlib
import logging as lg
import sys
from pathlib import Path


def set_path(path_dir_list: list[Path]):
    for path_dir in path_dir_list:
        r_path_dir = Path(path_dir).resolve()
        lg.info(f'Adding path: {r_path_dir}')
        sys.path.append(str(r_path_dir))


def get_cluster(base_dir: Path, cluster_package: str):
    r_base_dir = base_dir.resolve()
    lg.info(f'Using base: {r_base_dir}')
    sys.path.append(str(r_base_dir))
    lg.debug(f'Set Python path: {sys.path}')
    models = importlib.import_module(f'{cluster_package}.models')

    # NB: Reload is needed to handle the same file structure for multiple clusters
    importlib.reload(models)
    lg.debug(f'Removing base: {r_base_dir}')
    sys.path.remove(str(r_base_dir))

    for name in dir(models):
        obj = getattr(models, name)

        if isinstance(obj, models.Cluster):
            lg.info(f'Cluster found: {obj.prefix}')
            lg.debug(f'Cluster models: {obj.models}')
            return obj

    raise UserWarning('Cluster not found')



class Cluster:
    prefix: str
    models: dict[str, dict]
    exports: dict[str, dict]

    def __init__(self, prefix=''):
        self.prefix = prefix
        self.models = {}
        self.exports = {}

    def model(
        self,
        *,
        PK: str,
        SK: str | None = None,
        table: str | None = None,
        name: str | None = None,
        alias: str | None = None
    ):
        def decorator(cls):
            self.models[name or cls.__name__] = {
                'cls': cls,
                'main': {'PK': PK, 'SK': SK},
                'table': table or self.prefix + (name or cls.__name__),
                'indexes': {},
                'class_name': cls.__name__
            }
            return cls

        return decorator

    def index(self, *, index: str, PK: str, SK: str | None = None):
        def decorator(cls):
            if not self.models.get(cls.__name__):
                raise Exception(
                    'Model not found. Index decorator must be placed above model decorator'
                )

            self.models[cls.__name__]['indexes'][index] = {'PK': PK, 'SK': SK}
            return cls

        return decorator

    def export(self, cls):
        self.exports[cls.__name__] = {'cls': cls, 'class_name': cls.__name__}
        return cls



import re
from pathlib import Path
import logging as lg

import click
from mako.template import Template
from caseconverter import snakecase, pascalcase
import inspect
import gray_formatter

from prismarine.prisma_common import get_cluster


HEADER = '''\
# This file is generated by prismarine. Do not modify it.

from prismarine.runtime.dynamo_crud import (
    _query,
    _get_item,
    _update,
    _put_item,
    _delete,
    _scan,
    _save,
    Model,
    without,
    DbNotFound
)

from typing import TypedDict, List, Literal, NotRequired  # noqa
from types import EllipsisType

from {access_module} import get_dynamo_access
'''

DYNAMO_ACCESS = '''\
dynamo = get_dynamo_access()
'''


def build_client(cluster, base_dir: Path, runtime: str, access_module: str | None):
    r_base_dir = base_dir.resolve()
    this_dir = Path(__file__).resolve().parent
    template_file = Path(this_dir, 'model.mako')
    model_template = Template(template_file.read_text())
    module_body = ''
    imports = set()

    for name, model in cluster.models.items():
        file_path = Path(inspect.getfile(model['cls']))
        relative_arr = file_path.relative_to(r_base_dir).as_posix().split('/')
        relative_arr[-1] = relative_arr[-1].replace('.py', '')
        imports.add(('.'.join(relative_arr), model['class_name']))
        source = inspect.getsource(model['cls'])
        source = re.sub(r'@.+?\n', '', source)

        source = re.sub(
            r'[a-zA-Z\d]+\(TypedDict\)', 'UpdateDTO(TypedDict, total=False)', source
        )

        module_body += '\n\n\n' + str(
            model_template.render(
                snake=snakecase,
                pascal=pascalcase,
                ModelClass=f'{name}Model',
                Model=model['class_name'],
                PartitionKey=model['main']['PK'],
                SortKey=model['main']['SK'],
                Indexes=[
                    {
                        'name': index_name,
                        'PartitionKey': index['PK'],
                        'SortKey': index['SK']
                    }
                    for index_name, index in model['indexes'].items()
                ],
                TableName=model['table'],
                DtoLines=source.split('\n')
            )
        )

    # Additional exported classes (but not models)
    for export in cluster.exports.values():
        file_path = Path(inspect.getfile(export['cls']))
        relative_arr = file_path.relative_to(r_base_dir).as_posix().split('/')
        relative_arr[-1] = relative_arr[-1].replace('.py', '')
        imports.add(('.'.join(relative_arr), export['class_name']))

    access_module = access_module or 'prismarine.runtime.dynamo_default'
    header = HEADER.format(access_module=access_module)

    runtime_prefix = f'{runtime}.' if runtime else ''

    for i in imports:
        header += f'from {runtime_prefix}{i[0]} import {i[1]}\n'

    content = header + DYNAMO_ACCESS + module_body
    content = gray_formatter.fix_content(content)
    return content


def write_client(content, base_dir: Path, cluster_package: str):
    client_path = Path(base_dir, cluster_package, 'prismarine_client.py')
    lg.info('Writing ' + str(client_path))
    client_path.write_text(content)


@click.command()
@click.pass_obj
@click.argument('cluster_package')
def generate_client(obj, cluster_package):
    cluster = get_cluster(obj['BaseDir'], cluster_package)

    content = build_client(
        cluster, obj['BaseDir'], obj['Runtime'], obj['DynamoAccessModule']
    )

    write_client(content, obj['BaseDir'], cluster_package)
